"""Test Render_diff templatetags."""
import unittest
from nautobot_ssot.templatetags.render_diff import render_diff


test_params = [
    (
        {
            "region": {
                "Catalonia": {"+": {"parent_name": None}, "-": {"parent_name": "Europe"}},
            }
        },
        '<ul><li>region<ul><li class="diff-changed">Catalonia<ul><li class="diff-added">parent_name: None</li><li class="diff-subtracted">parent_name: Europe</li></ul></li></ul></li></ul>',
    ),
    (
        {
            "region": {
                "Barcelona": {
                    "+": {
                        "cfs": {"asw_owner": ""},
                        "slug": "barcelona",
                        "description": "",
                        "parent_name": "Catalonia",
                    }
                },
            }
        },
        '<ul><li>region<ul><li class="diff-added">Barcelona<ul><li class="diff-added">cfs: {&#x27;asw_owner&#x27;: &#x27;&#x27;}</li><li class="diff-added">slug: barcelona</li><li class="diff-added">description: </li><li class="diff-added">parent_name: Catalonia</li></ul></li></ul></li></ul>',
    ),
    (
        {
            "model_name": {
                "element": {
                    "-": {
                        "cfs": {"this is a XSS": "<script>alert(document.cookie)</script>"},
                    }
                },
            }
        },
        '<ul><li>model_name<ul><li class="diff-subtracted">element<ul><li class="diff-subtracted">cfs: {&#x27;this is a XSS&#x27;: &#x27;&lt;script&gt;alert(document.cookie)&lt;/script&gt;&#x27;}</li></ul></li></ul></li></ul>',
    ),
    (
        {
            "model_name": {
                "element": {
                    "-": {
                        "description": "<script>alert(document.cookie)</script>",
                    }
                },
            }
        },
        '<ul><li>model_name<ul><li class="diff-subtracted">element<ul><li class="diff-subtracted">description: &lt;script&gt;alert(document.cookie)&lt;/script&gt;</li></ul></li></ul></li></ul>',
    ),
]


class TestRenderDiff(unittest.TestCase):
    """Tests for render_diff function."""

    def test_render_diff_as_expected(self):
        """Testing expected escaped and rendered HTML."""
        for input_dict, rendered_diff in test_params:
            with self.subTest():
                self.assertEqual(render_diff(input_dict), rendered_diff)



"""Django urlpatterns declaration for nautobot_ssot plugin."""

from django.urls import path

from . import views

urlpatterns = [
    path("", views.DashboardView.as_view(), name="dashboard"),
    path("data-sources/<path:class_path>/", views.DataSourceTargetView.as_view(), name="data_source"),
    path("data-targets/<path:class_path>/", views.DataSourceTargetView.as_view(), name="data_target"),
    path("history/", views.SyncListView.as_view(), name="sync_list"),
    path("history/delete/", views.SyncBulkDeleteView.as_view(), name="sync_bulk_delete"),
    path("history/<uuid:pk>/", views.SyncView.as_view(), name="sync"),
    path("history/<uuid:pk>/delete/", views.SyncDeleteView.as_view(), name="sync_delete"),
    path("history/<uuid:pk>/jobresult/", views.SyncJobResultView.as_view(), name="sync_jobresult"),
    path("history/<uuid:pk>/logs/", views.SyncLogEntriesView.as_view(), name="sync_logentries"),
    path("logs/", views.SyncLogEntryListView.as_view(), name="synclogentry_list"),
]



"""Forms for working with Sync and SyncLogEntry models."""

from django import forms

from nautobot.utilities.forms import add_blank_choice, BootstrapMixin, BOOLEAN_WITH_BLANK_CHOICES

from .choices import SyncLogEntryActionChoices, SyncLogEntryStatusChoices
from .models import Sync, SyncLogEntry


class SyncFilterForm(BootstrapMixin, forms.ModelForm):
    """Form for filtering SyncOverview records."""

    dry_run = forms.ChoiceField(choices=BOOLEAN_WITH_BLANK_CHOICES, required=False)

    class Meta:
        """Metaclass attributes of SyncFilterForm."""

        model = Sync
        fields = ["dry_run"]


class SyncLogEntryFilterForm(BootstrapMixin, forms.ModelForm):
    """Form for filtering SyncLogEntry records."""

    q = forms.CharField(required=False, label="Search")
    sync = forms.ModelChoiceField(queryset=Sync.objects.all(), required=False)
    action = forms.ChoiceField(choices=add_blank_choice(SyncLogEntryActionChoices), required=False)
    status = forms.ChoiceField(choices=add_blank_choice(SyncLogEntryStatusChoices), required=False)

    class Meta:
        """Metaclass attributes of SyncLogEntryFilterForm."""

        model = SyncLogEntry
        fields = ["sync", "action", "status"]


class SyncForm(BootstrapMixin, forms.Form):
    """Base class for dynamic form generation for a SyncWorker."""

    dry_run = forms.BooleanField(
        required=False,
        initial=True,
        label="Dry run",
        help_text="Perform a dry run, making no actual changes to the database.",
    )



"""Plugin provision of Nautobot Job subclasses."""

from django.conf import settings

from nautobot.extras.jobs import get_jobs

from .base import DataSource, DataTarget
from .examples import ExampleDataSource, ExampleDataTarget

if settings.PLUGINS_CONFIG["nautobot_ssot"]["hide_example_jobs"]:
    jobs = []
else:
    jobs = [ExampleDataSource, ExampleDataTarget]


def get_data_jobs():
    """Get all data-source and data-target jobs available."""
    jobs_dict = get_jobs()
    data_sources = []
    data_targets = []
    for modules in jobs_dict.values():
        for module_data in modules.values():
            for job_class in module_data["jobs"].values():
                if issubclass(job_class, DataSource):
                    data_sources.append(job_class)
                if issubclass(job_class, DataTarget):
                    data_targets.append(job_class)

    return (data_sources, data_targets)



# Generated by Django 3.1.12 on 2021-06-28 20:03

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("extras", "0005_configcontext_device_types"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Sync",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True
                    ),
                ),
                ("source", models.CharField(max_length=64)),
                ("target", models.CharField(max_length=64)),
                ("start_time", models.DateTimeField(blank=True, null=True)),
                ("dry_run", models.BooleanField(default=False)),
                ("diff", models.JSONField(blank=True)),
                (
                    "job_result",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to="extras.jobresult"
                    ),
                ),
            ],
            options={
                "ordering": ["start_time"],
            },
        ),
        migrations.CreateModel(
            name="SyncLogEntry",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now_add=True)),
                ("action", models.CharField(max_length=32)),
                ("status", models.CharField(max_length=32)),
                ("diff", models.JSONField(blank=True, null=True)),
                ("synced_object_id", models.UUIDField(blank=True, null=True)),
                ("object_repr", models.CharField(blank=True, default="", editable=False, max_length=200)),
                ("message", models.CharField(blank=True, max_length=511)),
                (
                    "sync",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="logs",
                        related_query_name="log",
                        to="nautobot_ssot.sync",
                    ),
                ),
                (
                    "synced_object_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "sync log entries",
                "ordering": ["sync", "timestamp"],
            },
        ),
    ]



"""
Django Models for recording the status and progress of data synchronization between data sources.

The interaction between these models and Nautobot's native JobResult model deserves some examination.

- A JobResult is created each time a data sync is requested.
  - This stores a reference to the specific sync operation requested (JobResult.name),
    much as a Job-related JobResult would reference the name of the Job.
  - This stores a 'job_id', which this plugin uses to reference the specific sync instance.
  - This stores the 'created' and 'completed' timestamps, and the requesting user (if any)
  - This stores the overall 'status' of the job (pending, running, completed, failed, errored.)
  - This stores a 'data' field which, in theory can store arbitrary JSON data, but in practice
    expects a fairly strict structure for logging of various status messages.
    This field is therefore not suitable for storage of in-depth data synchronization log messages,
    which have a different set of content requirements, but is used for high-level status reporting.

JobResult 1<->1 Sync 1-->n SyncLogEntry
"""
from datetime import timedelta

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse
from django.utils.formats import date_format
from django.utils.timezone import now

from nautobot.core.models import BaseModel
from nautobot.extras.models import JobResult
from nautobot.extras.utils import extras_features

from .choices import SyncLogEntryActionChoices, SyncLogEntryStatusChoices


@extras_features(
    "custom_links",
)
class Sync(BaseModel):
    """High-level overview of a data sync event/process/attempt.

    Essentially an extension of the JobResult model to add a few additional fields.
    """

    source = models.CharField(max_length=64, help_text="System data is read from")
    target = models.CharField(max_length=64, help_text="System data is written to")

    start_time = models.DateTimeField(blank=True, null=True)
    # end_time is represented by the job_result.completed field
    source_load_time = models.DurationField(blank=True, null=True)
    target_load_time = models.DurationField(blank=True, null=True)
    diff_time = models.DurationField(blank=True, null=True)
    sync_time = models.DurationField(blank=True, null=True)
    source_load_memory_final = models.PositiveBigIntegerField(blank=True, null=True)
    source_load_memory_peak = models.PositiveBigIntegerField(blank=True, null=True)
    target_load_memory_final = models.PositiveBigIntegerField(blank=True, null=True)
    target_load_memory_peak = models.PositiveBigIntegerField(blank=True, null=True)
    diff_memory_final = models.PositiveBigIntegerField(blank=True, null=True)
    diff_memory_peak = models.PositiveBigIntegerField(blank=True, null=True)
    sync_memory_final = models.PositiveBigIntegerField(blank=True, null=True)
    sync_memory_peak = models.PositiveBigIntegerField(blank=True, null=True)

    dry_run = models.BooleanField(
        default=False, help_text="Report what data would be synced but do not make any changes"
    )
    diff = models.JSONField(blank=True)

    job_result = models.ForeignKey(to=JobResult, on_delete=models.PROTECT, blank=True, null=True)

    class Meta:
        """Metaclass attributes of Sync model."""

        ordering = ["start_time"]

    def __str__(self):
        """String representation of a Sync instance."""
        return f"{self.source} → {self.target}, {date_format(self.start_time, format=settings.SHORT_DATETIME_FORMAT)}"

    def get_absolute_url(self):
        """Get the detail-view URL for this instance."""
        return reverse("plugins:nautobot_ssot:sync", kwargs={"pk": self.pk})

    @classmethod
    def annotated_queryset(cls):
        """Construct an efficient queryset for this model and related data."""
        return (
            cls.objects.defer("diff")
            .select_related("job_result")
            .prefetch_related("logs")
            .annotate(
                num_unchanged=models.Count(
                    "log", filter=models.Q(log__action=SyncLogEntryActionChoices.ACTION_NO_CHANGE)
                ),
                num_created=models.Count("log", filter=models.Q(log__action=SyncLogEntryActionChoices.ACTION_CREATE)),
                num_updated=models.Count("log", filter=models.Q(log__action=SyncLogEntryActionChoices.ACTION_UPDATE)),
                num_deleted=models.Count("log", filter=models.Q(log__action=SyncLogEntryActionChoices.ACTION_DELETE)),
                num_succeeded=models.Count(
                    "log", filter=models.Q(log__status=SyncLogEntryStatusChoices.STATUS_SUCCESS)
                ),
                num_failed=models.Count("log", filter=models.Q(log__status=SyncLogEntryStatusChoices.STATUS_FAILURE)),
                num_errored=models.Count("log", filter=models.Q(log__status=SyncLogEntryStatusChoices.STATUS_ERROR)),
            )
        )

    @property
    def duration(self):
        """Total execution time of this Sync."""
        if not self.start_time:
            return timedelta()  # zero
        if not self.job_result or not self.job_result.completed:
            return now() - self.start_time
        return self.job_result.completed - self.start_time

    def get_source_url(self):
        """Get the absolute url of the source worker associated with this instance."""
        if self.source == "Nautobot" or not self.job_result:
            return None
        return reverse(
            "plugins:nautobot_ssot:data_source",
            kwargs={"class_path": self.job_result.name},
        )

    def get_target_url(self):
        """Get the absolute url of the target worker associated with this instance."""
        if self.target == "Nautobot" or not self.job_result:
            return None
        return reverse(
            "plugins:nautobot_ssot:data_target",
            kwargs={"class_path": self.job_result.name},
        )


class SyncLogEntry(BaseModel):
    """Record of a single event during a data sync operation.

    Detailed sync logs are recorded in this model, rather than in JobResult.data, because
    JobResult.data imposes fairly strict expectations about the structure of its contents
    that do not align well with the requirements of this plugin. Also, storing log entries as individual
    database records rather than a single JSON blob allows us to filter, query, sort, etc. as desired.

    This model somewhat "shadows" Nautobot's built-in ObjectChange model; the key distinction to
    bear in mind is that an ObjectChange reflects a change that *did happen*, while a SyncLogEntry
    may reflect this or may reflect a change that *could not happen* or *failed*.
    Additionally, if we're syncing data from Nautobot to a different system as data target,
    the data isn't changing in Nautobot, so there will be no ObjectChange record.
    """

    sync = models.ForeignKey(to=Sync, on_delete=models.CASCADE, related_name="logs", related_query_name="log")
    timestamp = models.DateTimeField(auto_now_add=True)

    action = models.CharField(max_length=32, choices=SyncLogEntryActionChoices)
    status = models.CharField(max_length=32, choices=SyncLogEntryStatusChoices)
    diff = models.JSONField(blank=True, null=True)

    synced_object_type = models.ForeignKey(
        to=ContentType,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
    )
    synced_object_id = models.UUIDField(blank=True, null=True)
    synced_object = GenericForeignKey(ct_field="synced_object_type", fk_field="synced_object_id")

    object_repr = models.TextField(blank=True, default="", editable=False)

    message = models.TextField(blank=True)

    class Meta:
        """Metaclass attributes of SyncLogEntry."""

        verbose_name_plural = "sync log entries"
        ordering = ["sync", "timestamp"]

    def get_action_class(self):
        """Map self.action to a Bootstrap label class."""
        return {
            SyncLogEntryActionChoices.ACTION_NO_CHANGE: "default",
            SyncLogEntryActionChoices.ACTION_CREATE: "success",
            SyncLogEntryActionChoices.ACTION_UPDATE: "info",
            SyncLogEntryActionChoices.ACTION_DELETE: "warning",
        }.get(self.action)

    def get_status_class(self):
        """Map self.status to a Bootstrap label class."""
        return {
            SyncLogEntryStatusChoices.STATUS_SUCCESS: "success",
            SyncLogEntryStatusChoices.STATUS_FAILURE: "warning",
            SyncLogEntryStatusChoices.STATUS_ERROR: "danger",
        }.get(self.status)



"""Simple template tag to render a shorter representation of a timedelta object."""

from django import template


register = template.Library()


@register.filter()
def humanize_bytes(size):
    """Humanize memory size given in bytes.

    Examples:
        1 => "1 B"
        1024 => "1 KiB"
        1024*1024 => "1 MiB"
    """
    if not isinstance(size, (int, float)):
        return "no data"

    suffix = "B"
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(size) < 1024.0:
            if round(size, 2) % 1 < 0.01:
                # If there are no relevant decimals "0.00" we skip them
                return f"{size:3.0f} {unit}{suffix}"
            return f"{size:3.2f} {unit}{suffix}"
        size /= 1024.0
    return f"{size:.1f} Yi{suffix}"



"""Basic tests that do not require Django."""
import unittest
import os
import toml

from nautobot_ssot import __version__ as project_version


class TestVersion(unittest.TestCase):
    """Test Version is the same."""

    def test_version(self):
        """Verify that pyproject.toml version is same as version specified in the package."""
        parent_path = os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
        poetry_version = toml.load(os.path.join(parent_path, "pyproject.toml"))["tool"]["poetry"]["version"]
        self.assertEqual(project_version, poetry_version)



"""Test the Job classes in nautobot_ssot."""
import os.path
from unittest.mock import Mock
import uuid
from django.contrib.contenttypes.models import ContentType

from django.forms import HiddenInput
from django.test import override_settings

# from django.test import TestCase

from nautobot.extras.models import JobResult
from nautobot.utilities.testing import TransactionTestCase

from nautobot_ssot.choices import SyncLogEntryActionChoices, SyncLogEntryStatusChoices
from nautobot_ssot.tests.jobs import DataSyncBaseJob, DataSource, DataTarget
from nautobot_ssot.models import SyncLogEntry


@override_settings(JOBS_ROOT=os.path.join(os.path.dirname(__file__), "jobs"))
class BaseJobTestCase(TransactionTestCase):
    """Test the DataSyncBaseJob class."""

    job_class = DataSyncBaseJob
    databases = (
        "default",
        "job_logs",
    )

    def setUp(self):
        """Per-test setup."""
        self.job = self.job_class()

        self.job.job_result = JobResult.objects.create(
            name="fake job",
            obj_type=ContentType.objects.get(app_label="extras", model="job"),
            job_id=uuid.uuid4(),
        )

        # name=self.job.class_path,

        self.job.load_source_adapter = lambda *x, **y: None
        self.job.load_target_adapter = lambda *x, **y: None

    def test_sync_log(self):
        """Test the sync_log() method."""
        self.job.run(data={"dry_run": True, "memory_profiling": False}, commit=True)
        self.assertIsNotNone(self.job.sync)
        # Minimal parameters
        self.job.sync_log(
            action=SyncLogEntryActionChoices.ACTION_CREATE,
            status=SyncLogEntryStatusChoices.STATUS_SUCCESS,
        )
        # Maximal parameters
        self.job.sync_log(
            action=SyncLogEntryActionChoices.ACTION_DELETE,
            status=SyncLogEntryStatusChoices.STATUS_ERROR,
            message="Whoops!",
            diff={"-": {"everything": "goodbye"}},
            synced_object=None,
            object_repr="Nothing to delete",
        )

        self.assertEqual(2, SyncLogEntry.objects.count())

    def test_as_form(self):
        """Test the as_form() method."""
        form = self.job.as_form()
        # Dry run flag defaults to true unless configured otherwise
        self.assertTrue(form.fields["dry_run"].initial)
        # Commit field is hidden to reduce user confusion
        self.assertIsInstance(form.fields["_commit"].widget, HiddenInput)

    def test_data_source(self):
        """Test the data_source property."""
        self.assertEqual(self.job.data_source, self.job_class.__name__)

    def test_data_target(self):
        """Test the data_target property."""
        self.assertEqual(self.job.data_target, self.job_class.__name__)

    def test_data_source_icon(self):
        """Test the data_source_icon property."""
        self.assertIsNone(self.job.data_source_icon)

    def test_data_target_icon(self):
        """Test the data_target_icon property."""
        self.assertIsNone(self.job.data_target_icon)

    def test_run(self):
        """Test the run() method."""
        self.job.run(data={"dry_run": True, "memory_profiling": False}, commit=True)
        self.assertIsNotNone(self.job.sync)
        self.assertIsNotNone(self.job.sync.source_load_time)
        self.assertIsNotNone(self.job.sync.target_load_time)
        self.assertIsNotNone(self.job.sync.diff_time)
        self.assertIsNone(self.job.sync.sync_time)
        self.assertEqual(self.job.sync.source, self.job.data_source)
        self.assertEqual(self.job.sync.target, self.job.data_target)
        self.assertTrue(self.job.sync.dry_run)
        self.assertEqual(self.job.job_result, self.job.sync.job_result)

    def test_calculate_diff(self):
        """Test calculate_diff() method."""
        self.job.sync = Mock()
        self.job.source_adapter = Mock()
        self.job.target_adapter = Mock()
        self.job.source_adapter.diff_to().dict.return_value = {}
        self.job.calculate_diff()
        self.job.source_adapter.diff_to.assert_called()
        self.job.sync.save.assert_called_once()


class DataSourceTestCase(BaseJobTestCase):
    """Test the DataSource class."""

    job_class = DataSource

    def test_data_target(self):
        """Test the override of the data_target property."""
        self.assertEqual(self.job.data_target, "Nautobot")

    def test_data_target_icon(self):
        """Test the override of the data_target_icon property."""
        self.assertEqual(self.job.data_target_icon, "/static/img/nautobot_logo.png")


class DataTargetTestCase(BaseJobTestCase):
    """Test the DataTarget class."""

    job_class = DataTarget

    def test_data_source(self):
        """Test the override of the data_source property."""
        self.assertEqual(self.job.data_source, "Nautobot")

    def test_data_source_icon(self):
        """Test the override of the data_source_icon property."""
        self.assertEqual(self.job.data_source_icon, "/static/img/nautobot_logo.png")



"""Plugin declaration for nautobot_ssot."""
# flake8: noqa
try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata

__version__ = metadata.version(__name__)

from nautobot.extras.plugins import PluginConfig


class NautobotSSOTPluginConfig(PluginConfig):
    """Plugin configuration for the nautobot_ssot plugin."""

    name = "nautobot_ssot"
    verbose_name = "Single Source of Truth"
    version = __version__
    author = "Network to Code, LLC"
    description = "Nautobot Single Source of Truth"
    base_url = "ssot"
    required_settings = []
    min_version = "1.0.3"
    max_version = "1.9999"
    default_settings = {
        "hide_example_jobs": False,
    }
    caching_config = {}


config = NautobotSSOTPluginConfig  # pylint:disable=invalid-name



"""Unit tests for nautobot_ssot plugin."""

from django.conf import settings

if "job_logs" in settings.DATABASES:
    settings.DATABASES["job_logs"] = settings.DATABASES["job_logs"].copy()
    settings.DATABASES["job_logs"]["TEST"] = {"MIRROR": "default"}



"""Override of classes."""
from nautobot.extras.jobs import Job
from nautobot_ssot.jobs import DataSource as _DataSource, DataTarget as _DataTarget
from nautobot_ssot.jobs.base import DataSyncBaseJob as _DataSyncBaseJob


class DataSyncBaseJob(_DataSyncBaseJob, Job):  # pylint: disable=abstract-method
    """DataSyncBaseJob import of the sync base job and Nautobot Job."""


class DataSource(_DataSource, Job):  # pylint: disable=abstract-method
    """DataSource import of the DataSource job and Nautobot Job."""


class DataTarget(_DataTarget, Job):  # pylint: disable=abstract-method
    """DataSyncBaseJob import of the DataTarget job and Nautobot Job."""



"""Django views for Single Source of Truth (SSoT)."""

import pprint

from django.contrib.contenttypes.models import ContentType
from django.http import Http404
from django.shortcuts import get_object_or_404, render
from django.views.generic import View

from django_tables2 import RequestConfig

from nautobot.extras.jobs import get_job
from nautobot.core.views.generic import BulkDeleteView, ObjectDeleteView, ObjectListView, ObjectView
from nautobot.utilities.paginator import EnhancedPaginator
from nautobot.utilities.views import ContentTypePermissionRequiredMixin

from .filters import SyncFilterSet, SyncLogEntryFilterSet
from .forms import SyncFilterForm, SyncLogEntryFilterForm
from .jobs import get_data_jobs, DataSource, DataTarget
from .models import Sync, SyncLogEntry
from .tables import DashboardTable, SyncTable, SyncTableSingleSourceOrTarget, SyncLogEntryTable


class DashboardView(ObjectListView):
    """Dashboard / overview of SSoT."""

    queryset = Sync.objects.all()
    table = DashboardTable
    action_buttons = []
    template_name = "nautobot_ssot/dashboard.html"

    def extra_context(self):
        """Extend the view context with additional details."""
        data_sources, data_targets = get_data_jobs()
        # Override default table context to limit the maximum number of records shown
        table = self.table(self.queryset, user=self.request.user)
        RequestConfig(
            self.request,
            {
                "paginator_class": EnhancedPaginator,
                "per_page": 10,
            },
        ).configure(table)
        context = {
            "queryset": self.queryset,
            "data_sources": data_sources,
            "data_targets": data_targets,
            "source": {},
            "target": {},
            "table": table,
        }
        sync_ct = ContentType.objects.get_for_model(Sync)
        for source in context["data_sources"]:
            context["source"][source.name] = self.queryset.filter(
                job_result__obj_type=sync_ct,
                job_result__name=source.class_path,
            )
        for target in context["data_targets"]:
            context["target"][target.name] = self.queryset.filter(
                job_result__obj_type=sync_ct,
                job_result__name=target.class_path,
            )

        return context


class DataSourceTargetView(ContentTypePermissionRequiredMixin, View):
    """Detail view of a given Data Source or Data Target Job."""

    additional_permissions = ("nautobot_ssot.view_sync",)

    def get_required_permission(self):
        """Permissions required to access this view."""
        return "extras.view_job"

    def get(self, request, class_path):
        """HTTP GET request handler."""
        job_class = get_job(class_path)
        if not job_class or not issubclass(job_class, (DataSource, DataTarget)):
            raise Http404

        syncs = Sync.annotated_queryset().filter(source=job_class.data_source, target=job_class.data_target)
        table = SyncTableSingleSourceOrTarget(syncs, user=request.user)

        return render(
            request,
            "nautobot_ssot/data_source_target.html",
            {
                "job_class": job_class,
                "table": table,
                "source_or_target": "source" if issubclass(job_class, DataSource) else "target",
            },
        )


class SyncListView(ObjectListView):
    """View for listing Sync records."""

    queryset = Sync.annotated_queryset()
    filterset = SyncFilterSet
    filterset_form = SyncFilterForm
    table = SyncTable
    action_buttons = []
    template_name = "nautobot_ssot/history.html"

    def extra_context(self):
        """Extend the view context with additional information."""
        data_sources, data_targets = get_data_jobs()
        return {
            "data_sources": data_sources,
            "data_targets": data_targets,
        }


class SyncDeleteView(ObjectDeleteView):
    """View for deleting a single Sync record."""

    queryset = Sync.objects.all()


class SyncBulkDeleteView(BulkDeleteView):
    """View for bulk-deleting Sync records."""

    queryset = Sync.objects.all()
    table = SyncTable


class SyncView(ObjectView):
    """View for details of a single Sync record."""

    queryset = Sync.annotated_queryset()
    template_name = "nautobot_ssot/sync_detail.html"

    def get_extra_context(self, request, instance):
        """Add additional context to the view."""
        return {
            "diff": pprint.pformat(instance.diff, width=180, compact=True),
        }


class SyncJobResultView(ObjectView):
    """View for the JobResult associated with a single Sync record."""

    queryset = Sync.objects.all()
    template_name = "nautobot_ssot/sync_jobresult.html"

    def get_extra_context(self, request, instance):
        """Add additional context to the view."""
        return {
            "active_tab": "jobresult",
        }


class SyncLogEntriesView(ObjectListView):
    """View for SyncLogEntries associated with a given Sync."""

    queryset = SyncLogEntry.objects.all()
    filterset = SyncLogEntryFilterSet
    filterset_form = SyncLogEntryFilterForm
    table = SyncLogEntryTable
    action_buttons = []
    template_name = "nautobot_ssot/sync_logentries.html"

    def get(self, request, pk):  # pylint: disable=arguments-differ
        """HTTP GET request handler."""
        self.instance = get_object_or_404(Sync.objects.all(), pk=pk)  # pylint: disable=attribute-defined-outside-init
        self.queryset = SyncLogEntry.objects.filter(sync=self.instance)

        return super().get(request)

    def extra_context(self):
        """Add additional context to the view."""
        return {"active_tab": "logentries", "object": self.instance}


class SyncLogEntryListView(ObjectListView):
    """View for listing SyncLogEntry records."""

    queryset = SyncLogEntry.objects.all()
    filterset = SyncLogEntryFilterSet
    filterset_form = SyncLogEntryFilterForm
    table = SyncLogEntryTable
    action_buttons = []
    template_name = "nautobot_ssot/synclogentry_list.html"



"""Template tag for rendering a DiffSync diff dictionary in a more human-readable form."""

from django import template
from django.utils.safestring import mark_safe
from django.utils.html import format_html


register = template.Library()


def render_diff_recursive(diff):
    """Recursively render a DiffSync diff dictionary representation to nested list elements.

    Example:
        {
            location:
                ams:
                    +: {name: "Amsterdam"}
                    -: {name: "amsterdam"}
                    device:
                        ams01:
                            +: {}
        }

        would render as:

        * location
          ! ams
            + name: Amsterdam
            - name: amsterdam
            * device
              + ams01
    """
    result = ""
    for record_type, children in diff.items():
        child_result = ""
        for child, child_diffs in children.items():
            if "+" in child_diffs and "-" not in child_diffs:
                child_class = "diff-added"
            elif "-" in child_diffs and "+" not in child_diffs:
                child_class = "diff-subtracted"
            elif not child_diffs.get("+") and not child_diffs.get("-"):
                child_class = "diff-unchanged"
            else:
                child_class = "diff-changed"
            child_result += format_html('<li class="{}">{}<ul>', child_class, child)

            for attr, value in child_diffs.pop("+", {}).items():
                child_result += format_html('<li class="diff-added">{}: {}</li>', attr, value)

            for attr, value in child_diffs.pop("-", {}).items():
                child_result += format_html('<li class="diff-subtracted">{}: {}</li>', attr, value)

            if child_diffs:
                child_result += render_diff_recursive(child_diffs)

            child_result += "</ul></li>"
        result += format_html("<li>{}<ul>{}</ul></li>", record_type, mark_safe(child_result))  # nosec
    return result


@register.simple_tag
def render_diff(diff):
    """Render a DiffSync diff dict to HTML."""
    html_text = render_diff_recursive(diff)
    return format_html("<ul>{}</ul>", mark_safe(html_text))  # nosec



"""ChoiceSet classes for Single Source of Truth (SSoT)."""

from nautobot.utilities.choices import ChoiceSet


class SyncLogEntryActionChoices(ChoiceSet):
    """Valid values for a SyncLogEntry.action field."""

    ACTION_NO_CHANGE = "no-change"
    ACTION_CREATE = "create"
    ACTION_UPDATE = "update"
    ACTION_DELETE = "delete"

    CHOICES = (
        (ACTION_NO_CHANGE, "no change"),
        (ACTION_CREATE, "create"),
        (ACTION_UPDATE, "update"),
        (ACTION_DELETE, "delete"),
    )


class SyncLogEntryStatusChoices(ChoiceSet):
    """Valid values for a SyncLogEntry.status field."""

    STATUS_SUCCESS = "success"
    STATUS_FAILURE = "failure"
    STATUS_ERROR = "error"

    CHOICES = (
        (STATUS_SUCCESS, "succeeded"),
        (STATUS_FAILURE, "failed"),
        (STATUS_ERROR, "errored"),
    )



"""Unit tests for nautobot_ssot."""
from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIClient

from nautobot.users.models import Token

User = get_user_model()


class PlaceholderAPITest(TestCase):
    """Test the nautobot_ssot API."""

    def setUp(self):
        """Create a superuser and token for API calls."""
        self.user = User.objects.create(username="testuser", is_superuser=True)
        self.token = Token.objects.create(user=self.user)
        self.client = APIClient()
        self.client.credentials(HTTP_AUTHORIZATION=f"Token {self.token.key}")

    def test_placeholder(self):
        """Verify that devices can be listed."""
        url = reverse("dcim-api:device-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["count"], 0)



"""View test cases for nautobot_ssot."""

from datetime import datetime
import uuid

from django.contrib.contenttypes.models import ContentType
from django.urls import reverse

from nautobot.extras.models import Job, JobResult
from nautobot.users.models import ObjectPermission
from nautobot.utilities.testing import ViewTestCases
from nautobot.utilities.testing.utils import disable_warnings

from nautobot_ssot.choices import SyncLogEntryActionChoices, SyncLogEntryStatusChoices
from nautobot_ssot.models import Sync, SyncLogEntry


class SyncViewsTestCase(  # pylint: disable=too-many-ancestors
    ViewTestCases.ListObjectsViewTestCase,
    ViewTestCases.GetObjectViewTestCase,
    ViewTestCases.DeleteObjectViewTestCase,
    ViewTestCases.BulkDeleteObjectsViewTestCase,
):
    """Test various views associated with the Sync model."""

    model = Sync

    @classmethod
    def setUpTestData(cls):
        """One-time setup of test data for this class."""
        for i in range(0, 3):
            job_result = JobResult.objects.create(
                name="plugins/nautobot_ssot.jobs.examples/ExampleDataSource",
                obj_type=ContentType.objects.get_for_model(Job),
                data={},
                job_id=uuid.uuid4(),
            )
            Sync.objects.create(
                source="Example Data Source",
                target="Nautobot",
                start_time=datetime.now(),
                dry_run=bool(i % 2),
                diff={},
                job_result=job_result,
            )

    def test_dashboard_without_permission(self):
        """Test that the dashboard view enforces permissions correctly."""
        with disable_warnings("django.request"):
            self.assertHttpStatus(self.client.get(reverse("plugins:nautobot_ssot:dashboard")), 403)

    def test_dashboard_with_permission(self):
        """Test the dashboard view."""
        obj_perm = ObjectPermission(name="Test permission", actions=["view"])
        obj_perm.save()
        obj_perm.users.add(self.user)
        obj_perm.object_types.add(ContentType.objects.get_for_model(self.model))

        self.assertHttpStatus(self.client.get(reverse("plugins:nautobot_ssot:dashboard")), 200)

    def test_data_source_target_view_without_permission(self):
        """Test that the DataSourceTargetView enforces permissions correctly."""
        with disable_warnings("django.request"):
            self.assertHttpStatus(
                self.client.get(
                    reverse(
                        "plugins:nautobot_ssot:data_source",
                        kwargs={"class_path": "plugins/nautobot_ssot.jobs.examples/ExampleDataSource"},
                    )
                ),
                403,
            )

        # Just access to Sync isn't sufficient - also need view_job permissions
        obj_perm = ObjectPermission(name="Test permission", actions=["view"])
        obj_perm.save()
        obj_perm.users.add(self.user)
        obj_perm.object_types.add(ContentType.objects.get_for_model(self.model))

        with disable_warnings("django.request"):
            self.assertHttpStatus(
                self.client.get(
                    reverse(
                        "plugins:nautobot_ssot:data_source",
                        kwargs={"class_path": "plugins/nautobot_ssot.jobs.examples/ExampleDataSource"},
                    )
                ),
                403,
            )

    def test_data_source_target_view_with_permission(self):
        """Test the DataSourceTargetView."""
        obj_perm = ObjectPermission(name="Test permission", actions=["view"])
        obj_perm.save()
        obj_perm.users.add(self.user)
        obj_perm.object_types.add(ContentType.objects.get_for_model(self.model))
        obj_perm.object_types.add(ContentType.objects.get_for_model(Job))
        self.assertHttpStatus(
            self.client.get(
                reverse(
                    "plugins:nautobot_ssot:data_source",
                    kwargs={"class_path": "plugins/nautobot_ssot.jobs.examples/ExampleDataSource"},
                )
            ),
            200,
        )

    def test_has_advanced_tab(self):
        pass


class SyncLogEntryViewsTestCase(ViewTestCases.ListObjectsViewTestCase):  # pylint: disable=too-many-ancestors
    """Test views related to the SyncLogEntry model."""

    model = SyncLogEntry

    @classmethod
    def setUpTestData(cls):
        """One-time setup of test data for this class."""
        job_result = JobResult.objects.create(
            name="plugins/nautobot_ssot.jobs.examples/ExampleDataSource",
            obj_type=ContentType.objects.get_for_model(Job),
            data={},
            job_id=uuid.uuid4(),
        )
        sync = Sync.objects.create(
            source="Example Data Source",
            target="Nautobot",
            start_time=datetime.now(),
            dry_run=False,
            diff={},
            job_result=job_result,
        )

        for i in range(0, 3):
            SyncLogEntry.objects.create(
                sync=sync,
                action=SyncLogEntryActionChoices.ACTION_NO_CHANGE,
                status=SyncLogEntryStatusChoices.STATUS_SUCCESS,
                diff={"+": {"i": i}},
                synced_object=None,
                object_repr="Placeholder",
                message="Log message",
            )

    def test_has_advanced_tab(self):
        pass



"""Plugin additions to the Nautobot navigation menu."""

from nautobot.extras.plugins import PluginMenuItem


menu_items = (
    PluginMenuItem(
        link="plugins:nautobot_ssot:dashboard",
        link_text="Dashboard",
        permissions=["nautobot_ssot.view_sync"],
    ),
    PluginMenuItem(
        link="plugins:nautobot_ssot:sync_list",
        link_text="History",
        permissions=["nautobot_ssot.view_sync"],
    ),
    PluginMenuItem(
        link="plugins:nautobot_ssot:synclogentry_list",
        link_text="Logs",
        permissions=["nautobot_ssot.view_synclogentry"],
    ),
)



"""Plugin template content extensions of base Nautobot views."""

from django.urls import reverse

from nautobot.extras.plugins import PluginTemplateExtension

from nautobot_ssot.models import Sync

# pylint: disable=abstract-method


class JobResultSyncLink(PluginTemplateExtension):
    """Add button linking to Sync data for relevant JobResults."""

    model = "extras.jobresult"

    def buttons(self):
        """Inject a custom button into the JobResult detail view, if applicable."""
        try:
            sync = Sync.objects.get(job_result=self.context["object"])
            return f"""
                <div class="btn-group">
                    <a href="{reverse('plugins:nautobot_ssot:sync', kwargs={'pk': sync.pk})}" class="btn btn-primary">
                        <span class="mdi mdi-database-sync-outline"></span> SSoT Sync Details
                    </a>
                </div>
            """
        except Sync.DoesNotExist:
            return ""


template_extensions = [JobResultSyncLink]



"""Sample data-source and data-target Jobs."""
# Skip colon check for multiple statements on one line.
# flake8: noqa: E701

from typing import Optional, Mapping
from uuid import UUID
from django.contrib.contenttypes.models import ContentType
from django.templatetags.static import static
from django.urls import reverse

from nautobot.dcim.models import Region, Site
from nautobot.ipam.models import Prefix
from nautobot.tenancy.models import Tenant
from nautobot.extras.jobs import Job, StringVar
from nautobot.extras.models import Status

from diffsync import DiffSync, DiffSyncModel
from diffsync.enum import DiffSyncFlags

import requests

from nautobot_ssot.jobs.base import DataMapping, DataSource, DataTarget


# In a more complex Job, you would probably want to move the DiffSyncModel subclasses into a separate Python module(s).

name = "SSoT Examples"  # pylint: disable=invalid-name


class RegionModel(DiffSyncModel):
    """Shared data model representing a Region in either of the local or remote Nautobot instances."""

    # Metadata about this model
    _modelname = "region"
    _identifiers = ("name",)
    _attributes = ("slug", "description", "parent_name")

    # Data type declarations for all identifiers and attributes
    name: str
    slug: str
    description: str
    parent_name: Optional[str]  # may be None

    # Not in _attributes or _identifiers, hence not included in diff calculations
    pk: Optional[UUID]


class SiteModel(DiffSyncModel):
    """Shared data model representing a Site in either of the local or remote Nautobot instances."""

    # Metadata about this model
    _modelname = "site"
    _identifiers = ("name",)
    # To keep this example simple, we don't include **all** attributes of a Site here. But you could!
    _attributes = ("slug", "status_slug", "region_name", "description")

    # Data type declarations for all identifiers and attributes
    name: str
    slug: str
    status_slug: str
    region_name: Optional[str]  # may be None
    description: str

    # Not in _attributes or _identifiers, hence not included in diff calculations
    pk: Optional[UUID]


class PrefixModel(DiffSyncModel):
    """Shared data model representing a Prefix in either of the local or remote Nautobot instances."""

    # Metadata about this model
    _modelname = "prefix"
    _identifiers = ("prefix", "tenant_slug")
    # To keep this example simple, we don't include **all** attributes of a Prefix here. But you could!
    _attributes = ("description", "status_slug")

    # Data type declarations for all identifiers and attributes
    prefix: str
    tenant_slug: str
    status_slug: str
    description: str

    # Not in _attributes or _identifiers, hence not included in diff calculations
    pk: Optional[UUID]


class RegionRemoteModel(RegionModel):
    """Implementation of Region create/update/delete methods for updating remote Nautobot data."""

    @classmethod
    def create(cls, diffsync, ids, attrs):
        """Create a new Region record in remote Nautobot.

        Args:
            diffsync (NautobotRemote): DiffSync adapter owning this Region
            ids (dict): Initial values for this model's _identifiers
            attrs (dict): Initial values for this model's _attributes
        """
        diffsync.post(
            "/api/dcim/regions/",
            {
                "name": ids["name"],
                "slug": attrs["slug"],
                "description": attrs["description"],
                "parent": {"name": attrs["parent_name"]} if attrs["parent_name"] else None,
            },
        )
        return super().create(diffsync, ids=ids, attrs=attrs)

    def update(self, attrs):
        """Update an existing Region record in remote Nautobot.

        Args:
            attrs (dict): Updated values for this record's _attributes
        """
        data = {}
        if "slug" in attrs:
            data["slug"] = attrs["slug"]
        if "description" in attrs:
            data["description"] = attrs["description"]
        if "parent_name" in attrs:
            if attrs["parent_name"]:
                data["parent"] = {"name": attrs["parent_name"]}
            else:
                data["parent"] = None
        self.diffsync.patch(f"/api/dcim/regions/{self.pk}/", data)
        return super().update(attrs)

    def delete(self):
        """Delete an existing Region record from remote Nautobot."""
        self.diffsync.delete(f"/api/dcim/regions/{self.pk}/")
        return super().delete()


class SiteRemoteModel(SiteModel):
    """Implementation of Site create/update/delete methods for updating remote Nautobot data."""

    @classmethod
    def create(cls, diffsync, ids, attrs):
        """Create a new Site in remote Nautobot.

        Args:
            diffsync (NautobotRemote): DiffSync adapter owning this Site
            ids (dict): Initial values for this model's _identifiers
            attrs (dict): Initial values for this model's _attributes
        """
        diffsync.post(
            "/api/dcim/sites/",
            {
                "name": ids["name"],
                "slug": attrs["slug"],
                "description": attrs["description"],
                "status": attrs["status_slug"],
                "region": {"name": attrs["region_name"]} if attrs["region_name"] else None,
            },
        )
        return super().create(diffsync, ids=ids, attrs=attrs)

    def update(self, attrs):
        """Update an existing Site record in remote Nautobot.

        Args:
            attrs (dict): Updated values for this record's _attributes
        """
        data = {}
        if "slug" in attrs:
            data["slug"] = attrs["slug"]
        if "description" in attrs:
            data["description"] = attrs["description"]
        if "status_slug" in attrs:
            data["status"] = attrs["status_slug"]
        if "region_name" in attrs:
            if attrs["region_name"]:
                data["region"] = {"name": attrs["region_name"]}
            else:
                data["region"] = None
        self.diffsync.patch(f"/api/dcim/sites/{self.pk}/", data)
        return super().update(attrs)

    def delete(self):
        """Delete an existing Site record from remote Nautobot."""
        self.diffsync.delete(f"/api/dcim/sites/{self.pk}/")
        return super().delete()


class PrefixRemoteModel(PrefixModel):
    """Implementation of Prefix create/update/delete methods for updating remote Nautobot data."""

    @classmethod
    def create(cls, diffsync, ids, attrs):
        """Create a new Prefix in remote Nautobot.

        Args:
            diffsync (NautobotRemote): DiffSync adapter owning this Prefix
            ids (dict): Initial values for this model's _identifiers
            attrs (dict): Initial values for this model's _attributes
        """
        diffsync.post(
            "/api/ipam/prefixes/",
            {
                "prefix": ids["prefix"],
                "tenant": {"slug": ids["tenant_slug"]} if ids["tenant_slug"] else None,
                "description": attrs["description"],
                "status": attrs["status_slug"],
            },
        )
        return super().create(diffsync, ids=ids, attrs=attrs)

    def update(self, attrs):
        """Update an existing Site record in remote Nautobot.

        Args:
            attrs (dict): Updated values for this record's _attributes
        """
        data = {}
        if "description" in attrs:
            data["description"] = attrs["description"]
        if "status_slug" in attrs:
            data["status"] = attrs["status_slug"]
        self.diffsync.patch(f"/api/dcim/sites/{self.pk}/", data)
        return super().update(attrs)

    def delete(self):
        """Delete an existing Site record from remote Nautobot."""
        self.diffsync.delete(f"/api/dcim/sites/{self.pk}/")
        return super().delete()


class RegionLocalModel(RegionModel):
    """Implementation of Region create/update/delete methods for updating local Nautobot data."""

    @classmethod
    def create(cls, diffsync, ids, attrs):
        """Create a new Region record in local Nautobot.

        Args:
            diffsync (NautobotLocal): DiffSync adapter owning this Region
            ids (dict): Initial values for this model's _identifiers
            attrs (dict): Initial values for this model's _attributes
        """
        region = Region(
            name=ids["name"],
            slug=attrs["slug"],
            description=attrs["description"],
        )
        if attrs["parent_name"]:
            region.parent = Region.objects.get(name=attrs["parent_name"])
        region.validated_save()
        return super().create(diffsync, ids=ids, attrs=attrs)

    def update(self, attrs):
        """Update an existing Region record in local Nautobot.

        Args:
            attrs (dict): Updated values for any of this model's _attributes
        """
        region = Region.objects.get(name=self.name)

        for attr_name in ("slug", "description"):
            if attr_name in attrs:
                setattr(region, attr_name, attrs[attr_name])

        if "parent_name" in attrs:
            region.parent = Region.objects.get(name=attrs["parent_name"])

        region.validated_save()
        return super().update(attrs)

    def delete(self):
        """Delete an existing Region record from local Nautobot."""
        region = Region.objects.get(name=self.name)
        region.delete()
        return super().delete()


class SiteLocalModel(SiteModel):
    """Implementation of Site create/update/delete methods for updating local Nautobot data."""

    @classmethod
    def create(cls, diffsync, ids, attrs):
        """Create a new Site record in local Nautobot.

        Args:
            diffsync (NautobotLocal): DiffSync adapter owning this Site
            ids (dict): Initial values for this model's _identifiers
            attrs (dict): Initial values for this model's _attributes
        """
        site = Site(name=ids["name"], slug=attrs["slug"], description=attrs["description"])
        site.status = Status.objects.get(slug=attrs["status_slug"])
        if attrs["region_name"]:
            site.region = Region.objects.get(name=attrs["region_name"])
        site.validated_save()
        return super().create(diffsync, ids=ids, attrs=attrs)

    def update(self, attrs):
        """Update an existing Site record in local Nautobot.

        Args:
            attrs (dict): Updated values for any of this model's _attributes
        """
        site = Site.objects.get(name=self.name)

        for attr_name in ("slug", "description"):
            if attr_name in attrs:
                setattr(site, attr_name, attrs[attr_name])

        if "status_slug" in attrs:
            site.status = Status.objects.get(slug=attrs["status_slug"])

        if "region_name" in attrs:
            site.region = Region.objects.get(name=attrs["region_name"])

        site.validated_save()
        return super().update(attrs)

    def delete(self):
        """Delete an existing Site record from local Nautobot."""
        site = Site.objects.get(name=self.name)
        site.delete()
        return super().delete()


class PrefixLocalModel(PrefixModel):
    """Implementation of Prefix create/update/delete methods for updating local Nautobot data."""

    @classmethod
    def create(cls, diffsync, ids, attrs):
        """Create a new Prefix record in local Nautobot.

        Args:
            diffsync (NautobotLocal): DiffSync adapter owning this Prefix
            ids (dict): Initial values for this model's _identifiers
            attrs (dict): Initial values for this model's _attributes
        """
        prefix = Prefix(prefix=ids["prefix"], description=attrs["description"])
        if ids["tenant_slug"]:
            tenant_obj, _ = Tenant.objects.get_or_create(
                slug=ids["tenant_slug"],
                defaults={"name": ids["tenant_slug"]},
            )
            prefix.tenant = tenant_obj

        status_obj, _ = Status.objects.get_or_create(
            slug=attrs["status_slug"],
            defaults={"name": attrs["status_slug"]},
        )
        status_obj.content_types.add(ContentType.objects.get_for_model(Prefix))
        prefix.status = status_obj
        prefix.validated_save()
        return super().create(diffsync, ids=ids, attrs=attrs)

    def update(self, attrs):
        """Update an existing Prefix record in local Nautobot.

        Args:
            attrs (dict): Updated values for any of this model's _attributes
        """
        prefix = Prefix.objects.get(prefix=self.prefix, tenant__slug=self.tenant_slug)
        for attr_name in ("description",):
            if attr_name in attrs:
                setattr(prefix, attr_name, attrs[attr_name])

        if "status_slug" in attrs:
            status_obj, _ = Status.objects.get_or_create(
                defaults={"name": attrs["status_slug"]},
            )
            status_obj.content_types.add(ContentType.objects.get_for_model(Prefix))
            prefix.status = status_obj
        prefix.validated_save()
        return super().update(attrs)

    def delete(self):
        """Delete an existing Prefix record from local Nautobot."""
        prefix = Prefix.objects.get(prefix=self.prefix, tenant__slug=self.tenant_slug)
        prefix.delete()
        return super().delete()


# In a more complex Job, you would probably want to move each DiffSync subclass into a separate Python module.


class NautobotRemote(DiffSync):
    """DiffSync adapter class for loading data from a remote Nautobot instance using Python requests.

    In a more realistic example, you'd probably use PyNautobot here instead of raw requests,
    but we didn't want to add PyNautobot as a dependency of this plugin just to make an example more realistic.
    """

    # Model classes used by this adapter class
    region = RegionRemoteModel
    site = SiteRemoteModel
    prefix = PrefixRemoteModel

    # Top-level class labels, i.e. those classes that are handled directly rather than as children of other models
    top_level = ("region", "site", "prefix")

    def __init__(self, *args, url=None, token=None, job=None, **kwargs):
        """Instantiate this class, but do not load data immediately from the remote system.

        Args:
            url (str): URL of the remote Nautobot system
            token (str): REST API authentication token
            job (Job): The running Job instance that owns this DiffSync adapter instance
        """
        super().__init__(*args, **kwargs)
        if not url or not token:
            raise ValueError("Both url and token must be specified!")
        self.url = url
        self.token = token
        self.job = job
        self.headers = {
            "Accept": "application/json",
            "Authorization": f"Token {self.token}",
        }

    def _get_api_data(self, url_path: str) -> Mapping:
        """Returns data from a url_path using pagination."""
        data = requests.get(f"{self.url}/{url_path}", headers=self.headers, params={"limit": 0}).json()
        result_data = data["results"]
        while data["next"]:
            data = requests.get(data["next"], headers=self.headers, params={"limit": 0}).json()
            result_data.extend(data["results"])
        return result_data

    def load(self):
        """Load Region and Site data from the remote Nautobot instance."""
        for region_entry in self._get_api_data("api/dcim/regions/"):
            region = self.region(
                name=region_entry["name"],
                slug=region_entry["slug"],
                description=region_entry["description"],
                parent_name=region_entry["parent"]["name"] if region_entry["parent"] else None,
                pk=region_entry["id"],
            )
            self.add(region)
            self.job.log_debug(message=f"Loaded {region} from remote Nautobot instance")

        for site_entry in self._get_api_data("api/dcim/sites/"):
            site = self.site(
                name=site_entry["name"],
                slug=site_entry["slug"],
                status_slug=site_entry["status"]["value"],
                region_name=site_entry["region"]["name"] if site_entry["region"] else None,
                description=site_entry["description"],
                pk=site_entry["id"],
            )
            self.add(site)
            self.job.log_debug(message=f"Loaded {site} from remote Nautobot instance")

        for prefix_entry in self._get_api_data("api/ipam/prefixes/"):
            prefix = self.prefix(
                prefix=prefix_entry["prefix"],
                description=prefix_entry["description"],
                status_slug=prefix_entry["status"]["value"],
                tenant_slug=prefix_entry["tenant"]["slug"] if prefix_entry["tenant"] else "",
                pk=prefix_entry["id"],
            )
            self.add(prefix)
            self.job.log_debug(message=f"Loaded {prefix} from remote Nautobot instance")

    def post(self, path, data):
        """Send an appropriately constructed HTTP POST request."""
        response = requests.post(f"{self.url}{path}", headers=self.headers, json=data)
        response.raise_for_status()
        return response

    def patch(self, path, data):
        """Send an appropriately constructed HTTP PATCH request."""
        response = requests.patch(f"{self.url}{path}", headers=self.headers, json=data)
        response.raise_for_status()
        return response

    def delete(self, path):
        """Send an appropriately constructed HTTP DELETE request."""
        response = requests.delete(f"{self.url}{path}", headers=self.headers)
        response.raise_for_status()
        return response


class NautobotLocal(DiffSync):
    """DiffSync adapter class for loading data from the local Nautobot instance."""

    # Model classes used by this adapter class
    region = RegionLocalModel
    site = SiteLocalModel
    prefix = PrefixLocalModel

    # Top-level class labels, i.e. those classes that are handled directly rather than as children of other models
    top_level = ("region", "site", "prefix")

    def __init__(self, *args, job=None, **kwargs):
        """Instantiate this class, but do not load data immediately from the local system."""
        super().__init__(*args, **kwargs)
        self.job = job

    def load(self):
        """Load Region and Site data from the local Nautobot instance."""
        for region in Region.objects.all():
            region_model = self.region(
                name=region.name,
                slug=region.slug,
                description=region.description,
                parent_name=region.parent.name if region.parent else None,
                pk=region.pk,
            )
            self.add(region_model)
            self.job.log_debug(message=f"Loaded {region_model} from local Nautobot instance")

        for site in Site.objects.all():
            site_model = self.site(
                name=site.name,
                slug=site.slug,
                status_slug=site.status.slug,
                region_name=site.region.name if site.region else None,
                description=site.description,
                pk=site.pk,
            )
            self.add(site_model)
            self.job.log_debug(message=f"Loaded {site_model} from local Nautobot instance")

        for prefix in Prefix.objects.all():
            prefix_model = self.prefix(
                prefix=str(prefix.prefix),
                description=prefix.description,
                status_slug=prefix.status.slug,
                tenant_slug=prefix.tenant.slug if prefix.tenant else "",
                pk=prefix.pk,
            )
            self.add(prefix_model)
            self.job.log_debug(message=f"Loaded {prefix_model} from local Nautobot instance")


# The actual Data Source and Data Target Jobs are relatively simple to implement
# once you have the above DiffSync scaffolding in place.


class ExampleDataSource(DataSource, Job):
    """Sync Region and Site data from a remote Nautobot instance into the local Nautobot instance."""

    source_url = StringVar(
        description="Remote Nautobot instance to load Sites and Regions from", default="https://demo.nautobot.com"
    )
    source_token = StringVar(description="REST API authentication token for remote Nautobot instance", default="a" * 40)

    def __init__(self):
        """Initialize ExampleDataSource."""
        super().__init__()
        self.diffsync_flags = self.diffsync_flags | DiffSyncFlags.SKIP_UNMATCHED_DST

    class Meta:
        """Metaclass attributes of ExampleDataSource."""

        name = "Example Data Source"
        description = 'Example "data source" Job for loading data into Nautobot from another system.'
        data_source = "Nautobot (remote)"
        data_source_icon = static("img/nautobot_logo.png")

    @classmethod
    def data_mappings(cls):
        """This Job maps Region and Site objects from the remote system to the local system."""
        return (
            DataMapping("Region (remote)", None, "Region (local)", reverse("dcim:region_list")),
            DataMapping("Site (remote)", None, "Site (local)", reverse("dcim:site_list")),
            DataMapping("Prefix (remote)", None, "Prefix (local)", reverse("ipam:prefix_list")),
        )

    def load_source_adapter(self):
        """Method to instantiate and load the SOURCE adapter into `self.source_adapter`."""
        self.source_adapter = NautobotRemote(url=self.kwargs["source_url"], token=self.kwargs["source_token"], job=self)
        self.source_adapter.load()

    def load_target_adapter(self):
        """Method to instantiate and load the TARGET adapter into `self.target_adapter`."""
        self.target_adapter = NautobotLocal(job=self)
        self.target_adapter.load()

    def lookup_object(self, model_name, unique_id):
        """Look up a Nautobot object based on the DiffSync model name and unique ID."""
        if model_name == "region":
            try:
                return Region.objects.get(name=unique_id)
            except Region.DoesNotExist:
                pass
        elif model_name == "site":
            try:
                return Site.objects.get(name=unique_id)
            except Site.DoesNotExist:
                pass
        elif model_name == "prefix":
            try:
                return Prefix.objects.get(
                    prefix=unique_id.split("__")[0], tenant__slug=unique_id.split("__")[1] or None
                )
            except Prefix.DoesNotExist:
                pass
        return None


class ExampleDataTarget(DataTarget, Job):
    """Sync Region and Site data from the local Nautobot instance to a remote Nautobot instance."""

    target_url = StringVar(description="Remote Nautobot instance to update", default="https://demo.nautobot.com")
    target_token = StringVar(description="REST API authentication token for remote Nautobot instance", default="a" * 40)

    def __init__(self):
        """Initialize ExampleDataTarget."""
        super().__init__()
        self.diffsync_flags = self.diffsync_flags | DiffSyncFlags.SKIP_UNMATCHED_DST

    class Meta:
        """Metaclass attributes of ExampleDataTarget."""

        name = "Example Data Target"
        description = 'Example "data target" Job for syncing data from Nautobot to another system'
        data_target = "Nautobot (remote)"
        data_target_icon = static("img/nautobot_logo.png")

    @classmethod
    def data_mappings(cls):
        """This Job maps Region and Site objects from the local system to the remote system."""
        return (
            DataMapping("Region (local)", reverse("dcim:region_list"), "Region (remote)", None),
            DataMapping("Site (local)", reverse("dcim:site_list"), "Site (remote)", None),
            DataMapping("Prefix (local)", reverse("ipam:prefix_list"), "Prefix (remote)", None),
        )

    def load_source_adapter(self):
        """Method to instantiate and load the SOURCE adapter into `self.source_adapter`."""
        self.source_adapter = NautobotLocal(job=self)
        self.source_adapter.load()

    def load_target_adapter(self):
        """Method to instantiate and load the TARGET adapter into `self.target_adapter`."""
        self.target_adapter = NautobotRemote(url=self.kwargs["target_url"], token=self.kwargs["target_token"], job=self)
        self.target_adapter.load()

    def lookup_object(self, model_name, unique_id):
        """Look up a Nautobot object based on the DiffSync model name and unique ID."""
        if model_name == "region":
            try:
                return Region.objects.get(name=unique_id)
            except Region.DoesNotExist:
                pass
        elif model_name == "site":
            try:
                return Site.objects.get(name=unique_id)
            except Site.DoesNotExist:
                pass
        elif model_name == "prefix":
            try:
                return Prefix.objects.get(
                    prefix=unique_id.split("__")[0], tenant__slug=unique_id.split("__")[1] or None
                )
            except Prefix.DoesNotExist:
                pass
        return None



"""Data tables for Single Source of Truth (SSOT) views."""
from django_tables2 import Column, DateTimeColumn, JSONColumn, LinkColumn, TemplateColumn

from nautobot.utilities.tables import BaseTable, ToggleColumn

from .choices import SyncLogEntryActionChoices, SyncLogEntryStatusChoices
from .models import Sync, SyncLogEntry


ACTION_LOGS_LINK = """
<a class="{{ link_class }}"
   href="{% url 'plugins:nautobot_ssot:synclogentry_list' %}?overview={{ record.id }}&action={{ action }}">
   {{ value }}
</a>
"""


STATUS_LOGS_LINK = """
<a class="{{ link_class }}"
   href="{% url 'plugins:nautobot_ssot:synclogentry_list' %}?overview={{ record.id }}&status={{ status }}">
   {{ value }}
</a>
"""


DRY_RUN_LABEL = """
{% if record.dry_run %}
<span class="dry_run label label-default">Dry Run</span>
{% else %}
<span class="dry_run label label-info">Sync</span>
{% endif %}
"""


MESSAGE_SPAN = """<span class="message">{% if record.message %}{{ record.message }}{% else %}—{% endif %}</span>"""


class DashboardTable(BaseTable):
    """Abbreviated version of SyncTable, for use with the dashboard."""

    start_time = DateTimeColumn(linkify=True, short=True)
    source = Column(linkify=lambda record: record.get_source_url())
    target = Column(linkify=lambda record: record.get_target_url())
    status = TemplateColumn(template_code="{% include 'extras/inc/job_label.html' with result=record.job_result %}")
    dry_run = TemplateColumn(template_code=DRY_RUN_LABEL, verbose_name="Type")

    class Meta(BaseTable.Meta):
        """Metaclass attributes of DashboardTable."""

        model = Sync
        fields = ["source", "target", "start_time", "status", "dry_run"]
        order_by = ["-start_time"]


class SyncTable(BaseTable):
    """Table for listing Sync records."""

    pk = ToggleColumn()
    source = Column(linkify=lambda record: record.get_source_url())
    target = Column(linkify=lambda record: record.get_target_url())
    start_time = DateTimeColumn(linkify=True, short=True)
    duration = TemplateColumn(template_code="{% load shorter_timedelta %}{{ record.duration | shorter_timedelta }}")
    dry_run = TemplateColumn(template_code=DRY_RUN_LABEL, verbose_name="Type")
    status = TemplateColumn(template_code="{% include 'extras/inc/job_label.html' with result=record.job_result %}")

    num_unchanged = TemplateColumn(
        template_code=ACTION_LOGS_LINK,
        verbose_name="No change",
        extra_context={"link_class": "num_unchanged", "action": SyncLogEntryActionChoices.ACTION_NO_CHANGE},
    )
    num_created = TemplateColumn(
        template_code=ACTION_LOGS_LINK,
        verbose_name="Create",
        extra_context={"link_class": "num_created", "action": SyncLogEntryActionChoices.ACTION_CREATE},
    )
    num_updated = TemplateColumn(
        template_code=ACTION_LOGS_LINK,
        verbose_name="Update",
        extra_context={"link_class": "num_updated", "action": SyncLogEntryActionChoices.ACTION_UPDATE},
    )
    num_deleted = TemplateColumn(
        template_code=ACTION_LOGS_LINK,
        verbose_name="Delete",
        extra_context={"link_class": "num_deleted", "action": SyncLogEntryActionChoices.ACTION_DELETE},
    )

    num_succeeded = TemplateColumn(
        template_code=STATUS_LOGS_LINK,
        verbose_name="Success",
        extra_context={"link_class": "num_succeeded", "status": SyncLogEntryStatusChoices.STATUS_SUCCESS},
    )
    num_failed = TemplateColumn(
        template_code=STATUS_LOGS_LINK,
        verbose_name="Failure",
        extra_context={"link_class": "num_failed", "status": SyncLogEntryStatusChoices.STATUS_FAILURE},
    )
    num_errored = TemplateColumn(
        template_code=STATUS_LOGS_LINK,
        verbose_name="Error",
        extra_context={"link_class": "num_errored", "status": SyncLogEntryStatusChoices.STATUS_ERROR},
    )

    class Meta(BaseTable.Meta):
        """Metaclass attributes of SyncTable."""

        model = Sync
        fields = (
            "pk",
            "source",
            "target",
            "start_time",
            "duration",
            "user",
            "status",
            "dry_run",
            "num_unchanged",
            "num_created",
            "num_updated",
            "num_deleted",
            "num_succeeded",
            "num_failed",
            "num_errored",
        )
        default_columns = (
            "pk",
            "source",
            "target",
            "start_time",
            "status",
            "dry_run",
            "num_created",
            "num_updated",
            "num_deleted",
            "num_failed",
            "num_errored",
        )
        order_by = ("-start_time",)


class SyncTableSingleSourceOrTarget(SyncTable):
    """Subclass of SyncTable with fewer default columns."""

    class Meta(SyncTable.Meta):
        """Metaclass attributes of SyncTableSingleSourceOrTarget."""

        default_columns = (
            "start_time",
            "status",
            "dry_run",
            "num_unchanged",
            "num_created",
            "num_updated",
            "num_deleted",
            "num_failed",
            "num_errored",
        )


ACTION_LABEL = """<span class="label label-{{ record.get_action_class }}">{{ record.action }}</span>"""


LOG_STATUS_LABEL = """<span class="label label-{{ record.get_status_class }}">{{ record.status }}</span>"""


SYNCED_OBJECT = """
{% if record.synced_object %}
<a href="{{ record.synced_object.get_absolute_url}}">{{ record.synced_object}}</a>
{% else %}
{{ record.object_repr }}
{% endif %}
"""


class SyncLogEntryTable(BaseTable):
    """Table for displaying SyncLogEntry records."""

    pk = ToggleColumn()
    sync = LinkColumn(verbose_name="Sync")
    action = TemplateColumn(template_code=ACTION_LABEL)
    status = TemplateColumn(template_code=LOG_STATUS_LABEL)
    diff = JSONColumn(orderable=False)
    message = TemplateColumn(template_code=MESSAGE_SPAN, orderable=False)
    synced_object = TemplateColumn(template_code=SYNCED_OBJECT)

    class Meta(BaseTable.Meta):
        """Metaclass attributes of SyncLogEntryTable."""

        model = SyncLogEntry
        fields = (
            "pk",
            "timestamp",
            "sync",
            "action",
            "synced_object_type",
            "synced_object",
            "status",
            "diff",
            "message",
        )
        default_columns = ("pk", "timestamp", "sync", "action", "synced_object", "status", "diff", "message")
        order_by = ("-timestamp",)



# Generated by Django 3.2.15 on 2022-09-27 13:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("nautobot_ssot", "0002_performance_metrics"),
    ]

    operations = [
        migrations.AlterField(
            model_name="synclogentry",
            name="message",
            field=models.TextField(blank=True),
        ),
        migrations.AlterField(
            model_name="synclogentry",
            name="object_repr",
            field=models.TextField(blank=True, default="", editable=False),
        ),
    ]



"""Model test cases for nautobot_ssot."""

from datetime import timedelta
import time
import uuid

from django.contrib.contenttypes.models import ContentType
from django.test import TestCase
from django.utils.timezone import now

from nautobot.extras.choices import JobResultStatusChoices
from nautobot.extras.models import Job, JobResult

from nautobot_ssot.models import Sync


class SyncTestCase(TestCase):
    """Tests for the Sync model."""

    def setUp(self):
        """Per-test setup function."""
        self.source_sync = Sync(
            source="Some other system",
            target="Nautobot",
            dry_run=False,
            start_time=None,
            diff={},
        )
        self.source_sync.validated_save()
        self.target_sync = Sync(
            source="Nautobot",
            target="Another system",
            dry_run=False,
            start_time=None,
            diff={},
        )
        self.target_sync.validated_save()

    def test_duration(self):
        """Test the duration property."""
        # Hasn't started yet, so no applicable duration
        self.assertEqual(self.source_sync.duration, timedelta())
        self.source_sync.start_time = now()
        time.sleep(1)
        self.assertGreater(self.source_sync.duration, timedelta())
        self.source_sync.job_result = JobResult(
            name="/plugins/nautobot_ssot.jobs.examples/ExampleDataSource",
            obj_type=ContentType.objects.get_for_model(Job),
            job_id=uuid.uuid4(),
        )
        # Still running
        time.sleep(1)
        self.assertGreater(self.source_sync.duration, timedelta(seconds=1))
        # Completed
        self.source_sync.job_result.set_status(JobResultStatusChoices.STATUS_COMPLETED)
        duration = self.source_sync.duration
        time.sleep(1)
        self.assertEqual(duration, self.source_sync.duration)

    def test_get_source_target_url(self):
        """Test the get_source_url() and get_target_url() methods."""
        # No JobResult
        self.assertIsNone(self.source_sync.get_source_url())
        self.assertIsNone(self.target_sync.get_target_url())
        # Source/target is Nautobot
        self.assertIsNone(self.target_sync.get_source_url())
        self.assertIsNone(self.source_sync.get_target_url())

        self.source_sync.job_result = JobResult(
            name="/plugins/nautobot_ssot.jobs.examples/ExampleDataSource",
            obj_type=ContentType.objects.get_for_model(Job),
            job_id=uuid.uuid4(),
        )
        self.target_sync.job_result = JobResult(
            name="/plugins/nautobot_ssot.jobs.examples/ExampleDataTarget",
            obj_type=ContentType.objects.get_for_model(Job),
            job_id=uuid.uuid4(),
        )

        self.assertIsNotNone(self.source_sync.get_source_url())
        self.assertIsNotNone(self.target_sync.get_target_url())
        # Source/target is Nautobot, so still None
        self.assertIsNone(self.target_sync.get_source_url())
        self.assertIsNone(self.source_sync.get_target_url())



# Generated by Django 3.1.13 on 2022-01-12 05:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("nautobot_ssot", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="sync",
            name="diff_memory_final",
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="diff_memory_peak",
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="diff_time",
            field=models.DurationField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="source_load_memory_final",
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="source_load_memory_peak",
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="source_load_time",
            field=models.DurationField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="sync_memory_final",
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="sync_memory_peak",
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="sync_time",
            field=models.DurationField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="target_load_memory_final",
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="target_load_memory_peak",
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="sync",
            name="target_load_time",
            field=models.DurationField(blank=True, null=True),
        ),
    ]



"""Filtering logic for Sync and SyncLogEntry records."""

import django_filters
from django.db.models import Q

from nautobot.utilities.filters import BaseFilterSet

from .models import Sync, SyncLogEntry


class SyncFilterSet(BaseFilterSet):
    """Filter capabilities for SyncOverview instances."""

    class Meta:
        """Metaclass attributes of SyncFilter."""

        model = Sync
        fields = ["dry_run", "job_result"]


class SyncLogEntryFilterSet(BaseFilterSet):
    """Filter capabilities for SyncLogEntry instances."""

    q = django_filters.CharFilter(method="search", label="Search")

    class Meta:
        """Metaclass attributes of SyncLogEntryFilter."""

        model = SyncLogEntry
        fields = ["sync", "action", "status", "synced_object_type"]

    def search(self, queryset, _name, value):  # pylint: disable=no-self-use
        """String search of SyncLogEntry records."""
        if not value.strip():
            return queryset
        return queryset.filter(Q(diff__icontains=value) | Q(message__icontains=value))



"""Helper templatetag for use with the "dashboard" UI view."""

import logging

from django import template


logger = logging.getLogger(__name__)


register = template.Library()


@register.inclusion_tag("nautobot_ssot/templatetags/dashboard_data.html")
def dashboard_data(sync_worker_class, queryset, kind="source"):
    """Render data about the sync history of a specific data-source or data-target."""
    if kind == "source":
        records = queryset.filter(source=sync_worker_class.data_source).order_by("-start_time")
    else:
        records = queryset.filter(target=sync_worker_class.data_target).order_by("-start_time")
    return {"syncs": records[:10], "count": records.count()}



"""Base Job classes for sync workers."""
from collections import namedtuple
from datetime import datetime
import traceback
import tracemalloc
from typing import Iterable
from packaging.version import Version

from django.forms import HiddenInput
from django.templatetags.static import static
from django.utils import timezone
from django.utils.functional import classproperty

# pylint-django doesn't understand classproperty, and complains unnecessarily. We disable this specific warning:
# pylint: disable=no-self-argument

from diffsync.enum import DiffSyncFlags
import structlog

from nautobot import __version__ as nautobot_version
from nautobot.extras.jobs import BaseJob, BooleanVar

from nautobot_ssot.choices import SyncLogEntryActionChoices
from nautobot_ssot.models import Sync, SyncLogEntry


DataMapping = namedtuple("DataMapping", ["source_name", "source_url", "target_name", "target_url"])
"""Entry in the list returned by a job's data_mappings() API.

The idea here is to provide insight into how various classes of data are mapped between Nautobot
and the other system acting as data source/target.

* source_name: Name of a data class (device, location, vlan, etc.) provided by the data source.
* source_url: URL (if any) to hyperlink for this source_name in the UI.
              Can be used, for example, to link to the Nautobot list view for this data class.
* target_name: Name of a data class on the data target that corresponds to the source data.
* target_url: URL (if any) to hyperlink the target_name.
"""


class DataSyncBaseJob(BaseJob):  # pylint: disable=too-many-instance-attributes
    """Common base class for data synchronization jobs.

    Works mostly as per the BaseJob API, with the following changes:

    - Concrete subclasses are responsible for implementing `self.sync_data()` (or related hooks), **not** `self.run()`.
    - Subclasses may optionally define any Meta field supported by Jobs, as well as the following:
      - `dry_run_default` - defaults to True if unspecified
      - `data_source` and `data_target` as labels (by default, will use the `name` and/or "Nautobot" as appropriate)
      - `data_source_icon` and `data_target_icon`
    """

    dry_run = BooleanVar()
    memory_profiling = BooleanVar(description="Perform a memory profiling analysis.", default=False)

    def load_source_adapter(self):
        """Method to instantiate and load the SOURCE adapter into `self.source_adapter`.

        Relevant available instance attributes include:

        - self.kwargs     (corresponds to the Job's `data` input, including 'dry_run' option)
        - self.job_result (as per Job API)
        """
        raise NotImplementedError

    def load_target_adapter(self):
        """Method to instantiate and load the TARGET adapter into `self.target_adapter`.

        Relevant available instance attributes include:

        - self.kwargs     (corresponds to the Job's `data` input, including 'dry_run' option)
        - self.job_result (as per Job API)
        """
        raise NotImplementedError

    def calculate_diff(self):
        """Method to calculate the difference from SOURCE to TARGET adapter and store in `self.diff`.

        This is a generic implementation that you could overwrite completely in your custom logic.
        """
        if self.source_adapter is not None and self.target_adapter is not None:
            self.diff = self.source_adapter.diff_to(self.target_adapter, flags=self.diffsync_flags)
            self.sync.diff = self.diff.dict()
            self.sync.save()
            self.log_info(message=self.diff.summary())
        else:
            self.log_warning(message="Not both adapters were properly initialized prior to diff calculation.")

    def execute_sync(self):
        """Method to synchronize the difference from `self.diff`, from SOURCE to TARGET adapter.

        This is a generic implementation that you could overwrite completely in your custom logic.
        """
        if self.source_adapter is not None and self.target_adapter is not None:
            self.source_adapter.sync_to(self.target_adapter, flags=self.diffsync_flags)
        else:
            self.log_warning(message="Not both adapters were properly initialized prior to synchronization.")

    def sync_data(self):
        """Method to load data from adapters, calculate diffs and sync (if not dry-run).

        It is composed by 4 methods:
        - self.load_source_adapter: instantiates the source adapter (self.source_adapter) and loads its data
        - self.load_target_adapter: instantiates the target adapter (self.target_adapter) and loads its data
        - self.calculate_diff: generates the diff from source to target adapter and stores it in self.diff
        - self.execute_sync: if not dry-run, uses the self.diff to synchronize from source to target

        This is a generic implementation that you could overwrite completely in you custom logic.
        Available instance attributes include:

        - self.kwargs     (corresponds to the Job's `data` input, including 'dry_run' option)
        - self.commit     (should generally be True)
        - self.sync       (Sync instance tracking this job execution)
        - self.job_result (as per Job API)
        """

        def record_memory_trace(step: str):
            """Helper function to record memory usage and reset tracemalloc stats."""
            memory_final, memory_peak = tracemalloc.get_traced_memory()
            setattr(self.sync, f"{step}_memory_final", memory_final)
            setattr(self.sync, f"{step}_memory_peak", memory_peak)
            self.sync.save()
            self.log_info(
                message=(f"Traced memory for {step} (Final, Peak): {memory_final} bytes, {memory_peak} bytes")
            )
            tracemalloc.clear_traces()

        if not self.sync:
            return

        if self.kwargs["memory_profiling"]:
            tracemalloc.start()

        start_time = datetime.now()

        self.log_info(message="Loading current data from source adapter...")
        self.load_source_adapter()
        load_source_adapter_time = datetime.now()
        self.sync.source_load_time = load_source_adapter_time - start_time
        self.sync.save()
        self.log_info(message=f"Source Load Time from {self.source_adapter}: {self.sync.source_load_time}")
        if self.kwargs["memory_profiling"]:
            record_memory_trace("source_load")

        self.log_info(message="Loading current data from target adapter...")
        self.load_target_adapter()
        load_target_adapter_time = datetime.now()
        self.sync.target_load_time = load_target_adapter_time - load_source_adapter_time
        self.sync.save()
        self.log_info(message=f"Target Load Time from {self.target_adapter}: {self.sync.target_load_time}")
        if self.kwargs["memory_profiling"]:
            record_memory_trace("target_load")

        self.log_info(message="Calculating diffs...")
        self.calculate_diff()
        calculate_diff_time = datetime.now()
        self.sync.diff_time = calculate_diff_time - load_target_adapter_time
        self.sync.save()
        self.log_info(message=f"Diff Calculation Time: {self.sync.diff_time}")
        if self.kwargs["memory_profiling"]:
            record_memory_trace("diff")

        if self.kwargs["dry_run"]:
            self.log_info("As `dry_run` is set, skipping the actual data sync.")
        else:
            self.log_info(message=f"Syncing from {self.source_adapter} to {self.target_adapter}...")
            self.execute_sync()
            execute_sync_time = datetime.now()
            self.sync.sync_time = execute_sync_time - calculate_diff_time
            self.sync.save()
            self.log_info(message="Sync complete")
            self.log_info(message=f"Sync Time: {self.sync.sync_time}")
            if self.kwargs["memory_profiling"]:
                record_memory_trace("sync")

    def lookup_object(self, model_name, unique_id):  # pylint: disable=no-self-use,unused-argument
        """Look up the Nautobot record, if any, identified by the args.

        Optional helper method used to build more detailed/accurate SyncLogEntry records from DiffSync logs.

        Args:
            model_name (str): DiffSyncModel class name or similar class/model label.
            unique_id (str): DiffSyncModel unique_id or similar unique identifier.

        Returns:
            models.Model: Nautobot model instance, or None
        """
        return None

    @classmethod
    def data_mappings(cls) -> Iterable[DataMapping]:
        """List the data mappings involved in this sync job."""
        return []

    @classmethod
    def config_information(cls):
        """Return a dict of user-facing configuration information {property: value}.

        Note that this will be rendered 'as-is' in the UI, so as a general practice this
        should NOT include sensitive information such as passwords!
        """
        return {}

    def sync_log(  # pylint: disable=too-many-arguments
        self,
        action,
        status,
        message="",
        diff=None,
        synced_object=None,
        object_repr="",
    ):
        """Log a action message as a SyncLogEntry."""
        if synced_object and not object_repr:
            object_repr = repr(synced_object)

        SyncLogEntry.objects.create(
            sync=self.sync,
            action=action,
            status=status,
            message=message,
            diff=diff,
            synced_object=synced_object,
            object_repr=object_repr,
        )

    def _structlog_to_sync_log_entry(self, _logger, _log_method, event_dict):
        """Capture certain structlog messages from DiffSync into the Nautobot database."""
        if all(key in event_dict for key in ("src", "dst", "action", "model", "unique_id", "diffs", "status")):
            # The DiffSync log gives us a model name (string) and unique_id (string).
            # Try to look up the actual Nautobot object that this describes.
            synced_object = self.lookup_object(  # pylint: disable=assignment-from-none
                event_dict["model"], event_dict["unique_id"]
            )
            object_repr = repr(synced_object) if synced_object else f"{event_dict['model']} {event_dict['unique_id']}"
            self.sync_log(
                action=event_dict["action"] or SyncLogEntryActionChoices.ACTION_NO_CHANGE,
                diff=event_dict["diffs"] if event_dict["action"] else None,
                status=event_dict["status"],
                message=event_dict["event"],
                synced_object=synced_object,
                object_repr=object_repr,
            )

        return event_dict

    @classmethod
    def _get_vars(cls):
        """Extend Job._get_vars to include `dry_run` variable.

        Workaround for https://github.com/netbox-community/netbox/issues/5529
        """
        got_vars = super()._get_vars()
        if hasattr(cls, "dry_run"):
            got_vars["dry_run"] = cls.dry_run

        if hasattr(cls, "memory_profiling"):
            got_vars["memory_profiling"] = cls.memory_profiling
        return got_vars

    def __init__(self):
        """Initialize a Job."""
        super().__init__()
        self.sync = None
        self.kwargs = {}
        self.commit = False
        self.diff = None
        self.source_adapter = None
        self.target_adapter = None
        # Default diffsync flags. You can overwrite them at any time.
        self.diffsync_flags = DiffSyncFlags.CONTINUE_ON_FAILURE | DiffSyncFlags.LOG_UNCHANGED_RECORDS

    def as_form(self, data=None, files=None, initial=None, approval_view=False):
        """Render this instance as a Django form for user inputs, including a "Dry run" field."""
        if Version(nautobot_version) < Version("1.2"):
            form = super().as_form(data=data, files=files, initial=initial)
        else:
            form = super().as_form(data=data, files=files, initial=initial, approval_view=approval_view)
        # Set the "dry_run" widget's initial value based on our Meta attribute, if any
        form.fields["dry_run"].initial = getattr(self.Meta, "dry_run_default", True)
        # Hide the "commit" widget to reduce user confusion
        form.fields["_commit"].widget = HiddenInput()
        return form

    @classproperty
    def data_source(cls):
        """The system or data source providing input data for this sync."""
        return getattr(cls.Meta, "data_source", cls.name)

    @classproperty
    def data_target(cls):
        """The system or data source being modified by this sync."""
        return getattr(cls.Meta, "data_target", cls.name)

    @classproperty
    def data_source_icon(cls):
        """Icon corresponding to the data_source."""
        return getattr(cls.Meta, "data_source_icon", None)

    @classproperty
    def data_target_icon(cls):
        """Icon corresponding to the data_target."""
        return getattr(cls.Meta, "data_target_icon", None)

    def run(self, data, commit):
        """Job entry point from Nautobot - do not override!"""
        self.sync = Sync.objects.create(
            source=self.data_source,
            target=self.data_target,
            dry_run=data["dry_run"],
            job_result=self.job_result,
            start_time=timezone.now(),
            diff={},
        )

        # Add _structlog_to_sync_log_entry as a processor for structlog calls from DiffSync
        structlog.configure(
            processors=[self._structlog_to_sync_log_entry, structlog.stdlib.render_to_log_kwargs],
            context_class=dict,
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )

        self.kwargs = data
        self.commit = commit

        # We need to catch exceptions and handle them, because if they aren't caught here,
        # they'll be caught by the Nautobot core run_job() function, which will trigger a database
        # rollback, which will delete our above created Sync record!
        try:
            self.sync_data()
        except Exception as exc:  # pylint: disable=broad-except
            stacktrace = traceback.format_exc()
            self.log_failure(message=f"An exception occurred: `{type(exc).__name__}: {exc}`\n```\n{stacktrace}\n```")


# pylint: disable=abstract-method
class DataSource(DataSyncBaseJob):
    """Base class for Jobs that sync data **from** another data source **to** Nautobot."""

    dry_run = BooleanVar(description="Perform a dry-run, making no actual changes to Nautobot data.")

    @classproperty
    def data_target(cls):
        """For a DataSource this is always Nautobot."""
        return "Nautobot"

    @classproperty
    def data_target_icon(cls):
        """For a DataSource this is always the Nautobot logo."""
        return static("img/nautobot_logo.png")


class DataTarget(DataSyncBaseJob):
    """Base class for Jobs that sync data **to** another data target **from** Nautobot."""

    dry_run = BooleanVar(description="Perform a dry-run, making no actual changes to the remote system.")

    @classproperty
    def data_source(cls):
        """For a DataTarget this is always Nautobot."""
        return "Nautobot"

    @classproperty
    def data_source_icon(cls):
        """For a DataTarget this is always the Nautobot logo."""
        return static("img/nautobot_logo.png")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nautobot_ssot',
 'nautobot_ssot.api',
 'nautobot_ssot.jobs',
 'nautobot_ssot.migrations',
 'nautobot_ssot.templatetags',
 'nautobot_ssot.tests',
 'nautobot_ssot.tests.jobs']

package_data = \
{'': ['*'],
 'nautobot_ssot': ['templates/nautobot_ssot/*',
                   'templates/nautobot_ssot/templatetags/*']}

install_requires = \
['Markdown!=3.3.5',
 'diffsync>=1.6.0,<2.0.0',
 'nautobot',
 'packaging>=21.3,<22.0']

entry_points = \
{'nautobot_ssot.data_sources': ['example = '
                                'nautobot_ssot.sync.example:ExampleSyncWorker'],
 'nautobot_ssot.data_targets': ['example = '
                                'nautobot_ssot.sync.example:ExampleSyncWorker']}

setup_kwargs = {
    'name': 'nautobot-ssot',
    'version': '1.2.0',
    'description': 'Nautobot Single Source of Truth',
    'long_description': '# Nautobot Single Source of Truth (SSoT)\n\nA plugin for [Nautobot](https://github.com/nautobot/nautobot). This plugin facilitates integration and data synchronization between various "source of truth" (SoT) systems, with Nautobot acting as a central clearinghouse for data - a Single Source of Truth, if you will.\n\nA goal of this plugin is to make it relatively quick and straightforward to [develop and integrate](https://nautobot-plugin-ssot.readthedocs.io/en/latest/developing_jobs/) your own system-specific Data Sources and Data Targets into Nautobot with a common UI and user experience.\n\n## Installation\n\nThe plugin is available as a Python package in PyPI and can be installed with `pip`:\n\n```shell\npip install nautobot-ssot\n```\n\n> This plugin is compatible with Nautobot 1.0.3 and higher.\n\nOnce installed, the plugin needs to be enabled in your `nautobot_config.py`:\n\n```python\n# In your nautobot_config.py\nPLUGINS = ["nautobot_ssot"]\n\nPLUGINS_CONFIG = {\n    "nautobot_ssot": {\n        "hide_example_jobs": False,  # defaults to False if unspecified\n    }\n}\n```\n\nThe plugin behavior can be controlled with the following list of settings:\n\n- `"hide_example_jobs"`: By default this plugin includes a pair of example data source / data target jobs so that you can see how it works without installing any additional plugins to provide specific system integrations. Once you have installed or developed some "real" system integrations to work with this plugin, you may wish to hide the example jobs, which you may do by setting this configuration setting to `True`.\n\n## Usage\n\nRefer to the [documentation](https://nautobot-plugin-ssot.readthedocs.io/en/latest/) for usage details.\n\n## Questions\n\nFor any questions or comments, please check the [FAQ](FAQ.md) first and feel free to swing by the [Network to Code slack channel](https://networktocode.slack.com/) (channel #networktocode).\nSign up [here](http://slack.networktocode.com/)\n\n## Screenshots\n\n![Dashboard screenshot](https://nautobot-plugin-ssot.readthedocs.io/en/latest/images/dashboard_initial.png)\n\n![Data Source detail view](https://nautobot-plugin-ssot.readthedocs.io/en/latest/images/data_source_detail.png)\n\n![Sync detail view](https://nautobot-plugin-ssot.readthedocs.io/en/latest/images/sync_detail.png)\n\n![Example data source - Arista CloudVision](https://nautobot-plugin-ssot.readthedocs.io/en/latest/images/example_cloudvision.png)\n\n![Example data target - ServiceNow](https://nautobot-plugin-ssot.readthedocs.io/en/latest/images/example_servicenow.png)\n',
    'author': 'Network to Code, LLC',
    'author_email': 'opensource@networktocode.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/nautobot/nautobot-plugin-ssot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Simple template tag to render a shorter representation of a timedelta object."""

from django import template
from django.utils.html import format_html


register = template.Library()


@register.filter
def shorter_timedelta(timedelta):
    """Render a timedelta as "HH:MM:SS.d" instead of "HH:MM:SS.mmmmmm".

    Note that we don't bother with rounding.
    """
    if timedelta:
        prefix, microseconds = str(timedelta).split(".", 1)
        deciseconds = int(microseconds) // 100_000
        return f"{prefix}.{deciseconds}"
    return format_html("&mdash;")
